"""
Sampling data from a 3D model
=============================

In this example we'll see how to sample a 3D model output at arbitrary points
within the model domain. This example uses a model with a single timestep,
so the time at which data is interpolated doesn't need to be specified.
"""
###############################################################################
# First, load the required modules.
import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np

from psipy.data import sample_data
from psipy.model import PLUTOOutput

###############################################################################
# Load a set of MAS output files, and get the number density variable from the
# model run. This example works with PLUTO data too. Uncomment the comment
# block below (and comment out the MAS block!)
"""
from psipy.model import MASOutput
mas_path = sample_data.mas_sample_data()
model = MASOutput(mas_path)
"""
pluto_path = sample_data.pluto_sample_data()
model = PLUTOOutput(pluto_path)
rho = model["rho"]

###############################################################################
# Choose a set of 1D points to interpolate the model output at.
#
# Here we keep a constant radius, and a set of longitudes that go all the way
# from 0 to 360 degrees. Then we choose two different, but close latitude
# values, and plot the results.
#
# As expected, the values at 0 and 360 degrees are the same, and the profiles
# are similar, but different, due to the small difference in latitude.
fig, ax = plt.subplots()

npoints = 1000
r = 50 * np.ones(npoints) * u.R_sun
lon = np.linspace(0, 360, npoints) * u.deg

for latitude in [0, 1] * u.deg:
    lat = latitude * np.ones(npoints)
    samples = rho.sample_at_coords(lon, lat, r)

    ax.plot(lon, samples, label="lat = " + str(latitude))

ax.legend()
ax.set_xlim(0, 360)
ax.set_ylim(bottom=0)
ax.set_xlabel("Longitude (deg)")
ax.set_ylabel(r"$\rho$ (cm$^{-3}$)")
ax.set_xticks([0, 90, 180, 270, 360])
ax.set_title(f"r={r[0].to_value(u.R_sun):.02f} R_sun")
plt.show()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

# -- Project information -----------------------------------------------------

project = "psipy"
copyright = "2020, Predictive Science Inc."
author = "David Stansby"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx_automodapi.automodapi",
    "sphinx_gallery.gen_gallery",
    "sphinx_rtd_theme",
    "numpydoc",
    "sphinx.ext.intersphinx",
]

numpydoc_show_class_members = False
sphinx_gallery_conf = {
    "examples_dirs": "../examples",
    "gallery_dirs": "auto_examples",
    "matplotlib_animations": True,
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Raise an error for missing doc references
nitpicky = True
nitpick_ignore = [
    ("py:class", "Unit('deg')"),
    ("py:class", "Unit('m')"),
    ("py:class", "Unit('rad')"),
]


intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "xarray": ("https://xarray.pydata.org/en/stable/", None),
    "astropy": ("https://docs.astropy.org/en/stable/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy", None),
    "heliopy": ("https://heliopy.readthedocs.io/en/0.15.4/", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "streamtracer": ("https://streamtracer.readthedocs.io/en/stable/", None),
    "pyvista": ("https://docs.pyvista.org/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
}

default_role = "any"
automodapi_inheritance_diagram = False
automodsumm_inherited_members = True



from dataclasses import dataclass
from typing import List

import astropy.units as u
import numpy as np
from astropy.coordinates import spherical_to_cartesian

__all__ = ["FieldLines", "FieldLine"]


@dataclass
class FieldLine:
    """
    A single field line.
    """

    r: np.ndarray
    lat: np.ndarray
    lon: np.ndarray

    def __init__(self, *, r, lat, lon):
        """
        Parameters
        ----------
        r : numpy.ndarray
            Radial coordinates.
        lat : numpy.ndarray
            Latitude coordinates **in radians**.
        lon : numpy.ndarray
            Longitude coordinates **in radians**.
        """
        self.r = r
        self.lat = lat * u.rad
        self.lon = lon * u.rad

    @property
    def xyz(self):
        """
        Cartesian coordinates as a (n, 3) shaped array.
        """
        return np.array(spherical_to_cartesian(self.r, self.lat, self.lon)).T

    @property
    def _rlatlon(self):
        """
        Spherical coordinates as a (n, 3) shaped array.
        """
        return np.column_stack(
            [self.r, self.lat.to_value(u.rad), self.lon.to_value(u.rad)]
        )


@dataclass
class FieldLines:
    """
    A container for multiple field lines.
    """

    flines: List[FieldLine]

    def __init__(self, xs):
        """
        Parameters
        ----------
        xs : list[numpy.ndarray]
            Field lines. Each array must have lon, lat, r columns in that
            order.
        """
        self.flines = [FieldLine(r=x[:, 2], lat=x[:, 1], lon=x[:, 0]) for x in xs]

    def __getitem__(self, i):
        return self.flines[i]

    def __len__(self):
        return len(self.flines)

    def __iter__(self):
        for fline in self.flines:
            yield fline

    def save(self, filename):
        """
        Save field lines to file.

        Parameters
        ----------
        filename : pathlib.Path, str
            File to save field lines to.

        Notes
        -----
        Arrays are saved using `numpy.savez_compressed`.
        """
        np.savez_compressed(filename, *[fline._rlatlon for fline in self.flines])

    @classmethod
    def load(cls, filename):
        """
        Load field lines from a file.

        The field lines must have been saved using the ``.save()`` method.

        Parameters
        ----------
        filename : pathlib.Path, str
            File to load field lines from.

        Returns
        -------
        flines : FieldLines
        """
        arrs = np.load(str(filename))
        return cls([arrs[k][:, ::-1] for k in arrs])



from psipy.data import sample_data
from psipy.model import MASOutput


def test_helio_timestamps():
    # Check that helio sample data has > 1 timesteps
    mas_path = sample_data.mas_helio_timesteps()
    model = MASOutput(mas_path)
    vr = model["vr"]
    assert vr.n_timesteps > 1



"""
This file includes global test configuration.

In particular, it defines the location where test data is available.
"""
from pathlib import Path

import pytest

from psipy.data import sample_data
from psipy.model import mas, pluto

test_data_dir = (Path(__file__) / ".." / ".." / "data").resolve()


@pytest.fixture(scope="module", params=["mas_helio", "mas_hdf5"])
def mas_directory(request):
    if request.param == "mas_helio":
        # Check for and download data if not present
        directory = sample_data.mas_sample_data(type="helio")
    else:
        # Directories with MAS outputs
        directory = test_data_dir / request.param
    if not directory.exists():
        pytest.xfail(f"Could not find MAS data directory at {directory}")

    return directory


@pytest.fixture(scope="module")
def mas_model(mas_directory):
    return mas.MASOutput(mas_directory)


@pytest.fixture(scope="module")
def pluto_directory():
    directory = sample_data.pluto_sample_data()
    if not directory.exists():
        pytest.xfail(f"Could not find PLUTO data directory at {directory}")

    return directory


@pytest.fixture(scope="module")
def pluto_model(pluto_directory):
    return pluto.PLUTOOutput(pluto_directory)



import astropy.units as u
import numpy as np
import pytest

from psipy.model import Variable


def test_var_error(mas_model):
    with pytest.raises(RuntimeError, match="not in list of known variables"):
        mas_model["not_a_var"]


def test_radial_normalised(mas_model):
    norm = mas_model["rho"].radial_normalized(-2)
    assert isinstance(norm, Variable)


# Check different shaped input, including lon/lat points that go up/down in
# value
@pytest.mark.parametrize(
    "lon, lat, r",
    [
        (1 * u.deg, 1 * u.deg, 30 * u.R_sun),
        ([1, 2] * u.deg, [1, 2] * u.deg, [30, 31] * u.R_sun),
        ([1, 0] * u.deg, [1, 0] * u.deg, [30, 31] * u.R_sun),
    ],
)
def test_sample_at_coords_mas(mas_model, lon, lat, r):
    # Check scalar coords
    rho = mas_model["rho"].sample_at_coords(lon=lon, lat=lat, r=r)
    assert rho.unit == mas_model["rho"].unit
    assert u.allclose(rho[0], [447.02795493] * u.cm**-3)


@pytest.mark.parametrize(
    "lon, lat, r",
    [
        (1 * u.deg, 1 * u.deg, 1 * u.AU),
        ([1, 2] * u.deg, [1, 2] * u.deg, [1, 1.01] * u.AU),
        ([1, 0] * u.deg, [1, 0] * u.deg, [1, 1.01] * u.AU),
    ],
)
def test_sample_at_coords_pluto(pluto_model, lon, lat, r):
    # Check scalar coords
    rho = pluto_model["rho"].sample_at_coords(lon=lon, lat=lat, r=r)
    assert rho.unit == pluto_model["rho"].unit
    assert u.allclose(rho[0], [13.50442343])


def test_sample_out_of_bounds(pluto_model):
    lon = [0, 0, 0] * u.deg
    lat = [0, 0, 0] * u.deg
    rho = pluto_model["rho"]
    # Check point below bounds, in bounds, above bounds
    r = [0, 0.5, 2] * u.AU
    assert r[0] < rho.r_coords[0]
    assert rho.r_coords[0] < r[1] < rho.r_coords[-1]
    assert rho.r_coords[-1] < r[2]
    # Check scalar coords
    with pytest.warns(UserWarning, match="outside bounds"):
        samples = rho.sample_at_coords(lon=lon, lat=lat, r=r)
    assert np.isnan(samples[0])
    assert not np.isnan(samples[1])
    assert np.isnan(samples[2])



import astropy.units as u
import numpy as np
import scipy.interpolate
import xarray as xr

from psipy.io import get_mas_variables, read_mas_file
from .base import ModelOutput

__all__ = ["MASOutput"]


# A mapping from unit names to their units, and factors the data needs to be
# multiplied to get them into these units.
_vunit = [u.km / u.s, 481.37]
_bunit = [u.G, 2.205]
_junit = [u.A / u.m**2, 2.267e4]
_mas_units = {
    "vr": _vunit,
    "vt": _vunit,
    "vp": _vunit,
    "va": _vunit,
    "br": _bunit,
    "bt": _bunit,
    "bp": _bunit,
    "bmag": _bunit,
    "rho": [u.cm**-3, 1.67e-16 / 1.67e-24],
    "t": [u.K, 2.804e7],
    "p": [u.Pa, 3.875717e-2],
    "jr": _junit,
    "jt": _junit,
    "jp": _junit,
}
_2pi = 2 * np.pi


class MASOutput(ModelOutput):
    """
    The results from a single run of MAS.

    This is a storage object that contains a number of `Variable` objects. It
    is designed to be used like::

        mas_output = MASOutput('directory')
        br = mas_output['br']

    Notes
    -----
    Variables are loaded on demand. To see the list of available variables
    use `MASOutput.variables`, and to see the list of already loaded variables
    use `MASOutput.loaded_variables`.
    """

    def get_unit(self, var):
        return _mas_units[var]

    def get_runit(self):
        return u.R_sun

    def get_variables(self):
        return get_mas_variables(self.path)

    def load_file(self, var):
        return read_mas_file(self.path, var)

    def __repr__(self):
        return f'psipy.model.mas.MASOutput("{self.path}")'

    def __str__(self):
        return f"MAS output in directory {self.path}\n" + super().__str__()

    def cell_corner_b(self, t_idx=None):
        """
        Get the magnetic field vector at the cell corners.

        Parameters
        ----------
        t_idx : int, optional
            If more than one timestep is present in the loaded model, a
            timestep index at which to get the vectors must be provided.

        Returns
        -------
        xarray.DataArray

        Notes
        -----
        The phi limits go from 0 to 2pi inclusive, with the vectors at phi=0
        equal to the vectors at phi=2pi.
        """
        if not set(["br", "bt", "bp"]) <= set(self.variables):
            raise RuntimeError("MAS output must have the br, bt, bp variables loaded")

        # Interpolate radial coordinate
        new_rcoord = self["bt"].r_coords
        br = scipy.interpolate.interp1d(
            self["br"].r_coords,
            self["br"].data.isel(time=t_idx or 0),
            axis=2,
            fill_value="extrapolate",
        )(new_rcoord)

        # Interpolate theta coordinate
        new_tcoord = self["bp"].theta_coords
        bt = scipy.interpolate.interp1d(
            self["bt"].theta_coords,
            self["bt"].data.isel(time=t_idx or 0),
            axis=1,
            fill_value="extrapolate",
        )(new_tcoord)

        # Interoplate phi coordinate
        new_pcoord = self["br"].phi_coords
        bp = scipy.interpolate.interp1d(
            self["bp"].phi_coords,
            self["bp"].data.isel(time=t_idx or 0),
            axis=0,
            fill_value="extrapolate",
        )(new_pcoord)
        # Calculate edge/cyclic phi value
        old_pcoord = self["bp"].phi_coords
        edge_pcoord = [old_pcoord[-1], old_pcoord[0] + _2pi]
        edge_data = self["bp"].data.isel(time=t_idx or 0)
        edge_data = np.stack([edge_data[-1, :, :], edge_data[0, :, :]], axis=0)
        bp_edge = scipy.interpolate.interp1d(edge_pcoord, edge_data, axis=0)(_2pi)
        bp_edge = bp_edge.reshape((1, *bp_edge.shape))

        # Add an extra layer of cells at phi=2pi for the tracer
        br = np.concatenate((br, br[0:1]), axis=0)
        bt = np.concatenate((bt, bt[0:1]), axis=0)
        bp = np.concatenate((bp_edge, bp[1:, :, :], bp_edge), axis=0)
        new_pcoord = np.append(new_pcoord, _2pi)

        return xr.DataArray(
            np.stack([bp, bt, br], axis=-1),
            dims=["phi", "theta", "r", "component"],
            coords=[new_pcoord, new_tcoord, new_rcoord, ["bp", "bt", "br"]],
        )

    def cell_centered_v(self, extra_phi_coord=False):
        """
        Get the velocity vector at the cell centres.

        Because the locations of the vector component outputs

        Parameters
        ----------
        extra_phi_coord: bool
            If `True`, add an extra phi slice.
        """
        if not set(["vr", "vt", "vp"]) <= set(self.variables):
            raise RuntimeError("MAS output must have the vr, vt, vp variables loaded")

        # Interpolate new radial coordiantes
        new_rcoord = self["vr"].r_coords
        vt = scipy.interpolate.interp1d(self["vt"].r_coords, self["vt"].data, axis=2)(
            new_rcoord
        )
        vp = scipy.interpolate.interp1d(self["vp"].r_coords, self["vp"].data, axis=2)(
            new_rcoord
        )

        # Interpolate new theta coordinates
        new_tcoord = self["vt"].theta_coords
        vr = scipy.interpolate.interp1d(
            self["vr"].theta_coords, self["vr"].data, axis=1
        )(new_tcoord)
        vp = scipy.interpolate.interp1d(self["vp"].theta_coords, vp, axis=1)(new_tcoord)
        # Don't need to interpolate phi coords, but get a copy
        new_pcoord = self["vr"].phi_coords

        if extra_phi_coord:
            dphi = np.mean(np.diff(new_pcoord))
            assert np.allclose(new_pcoord[0] + 2 * np.pi, new_pcoord[-1] + dphi)

            new_pcoord = np.append(new_pcoord, new_pcoord[-1] + dphi)
            vp = np.append(vp, vp[0:1, :, :], axis=0)
            vt = np.append(vt, vt[0:1, :, :], axis=0)
            vr = np.append(vr, vr[0:1, :, :], axis=0)

        return xr.DataArray(
            np.stack([vp, vt, vr], axis=-1),
            dims=["phi", "theta", "r", "component"],
            coords=[new_pcoord, new_tcoord, new_rcoord, ["vp", "vt", "vr"]],
        )



"""
Changing the units of a plot
=============================

This example shows how to change the radial coordiantes. This is helpful if you
want to change the radial coordinates e.g. from solar radii to AU.
"""
###############################################################################
# First, load the required modules.
import astropy.units as u
import matplotlib.pyplot as plt

from psipy.data import sample_data
from psipy.model import MASOutput

###############################################################################
# Next, load a set of MAS output files. You will need to change this line to
# point to a folder with MAS files in them.
mas_path = sample_data.mas_sample_data()
model = MASOutput(mas_path)

###############################################################################
# Convert from solar radii to AU

br = model["br"]
print(f"Old coords: {br.r_coords}")
br.r_coords = br.r_coords.to(u.AU)
print(f"New coords: {br.r_coords}")

###############################################################################
# Plot
cbar_kwargs = {"orientation": "horizontal"}
ax = plt.subplot(projection="polar")
br.plot_equatorial_cut(ax=ax, cbar_kwargs=cbar_kwargs)

plt.show()



import astropy.units as u
import numpy as np

from .flines import FieldLines

__all__ = ["FortranTracer"]


class FortranTracer:
    r"""
    Tracer using Fortran code.

    Parameters
    ----------
    max_steps: 'auto', int
        Maximum number of steps each streamline can take before stopping. This
        directly sets the memory allocated to the traced streamlines, so do not
        set it too large. If set to ``'auto'`` (the default),
    step_size : float
        Step size as a fraction of the smallest radial grid spacing.

    Notes
    -----
    Because the stream tracing is done in spherical coordinates, there is a
    singularity at the poles, which means seeds placed directly on the poles
    will not go anywhere.
    """

    def __init__(self, max_steps="auto", step_size=1):
        try:
            import streamtracer  # NoQA
        except ModuleNotFoundError as e:
            raise RuntimeError(
                "Using FortranTracer requires the streamtracer module, "
                "but streamtracer could not be loaded"
            ) from e
        self.max_steps = max_steps
        self.step_size = step_size
        self.max_steps = max_steps
        max_steps = 1 if max_steps == "auto" else max_steps
        # We have to set max_steps and step_size here to create a tracer,

    def _vector_grid(self, mas_output, t_idx):
        """
        Create a `streamtracer.VectorGrid` object from a MAS output.
        """
        bs = mas_output.cell_corner_b(t_idx)
        return self._vector_grid_from_bs(bs)

    def _vector_grid_from_bs(self, bs):
        """
        Create a `streamtracer.VectorGrid` object from a magnetic field array.
        """
        from streamtracer import VectorGrid

        # Account for tracing in spherical coordinates
        bs.loc[..., "bp"] /= np.cos(bs.coords["theta"])
        bs.loc[..., "bp"] /= bs.coords["r"]
        bs.loc[..., "bt"] /= bs.coords["r"]

        # cyclic only in the phi direction
        cyclic = [True, False, False]
        grid_coords = [
            bs.coords["phi"].values,
            bs.coords["theta"].values,
            bs.coords["r"].values,
        ]
        vector_grid = VectorGrid(bs.data, cyclic=cyclic, grid_coords=grid_coords)
        return vector_grid

    @u.quantity_input
    def trace(self, mas_output, *, r: u.m, lat: u.rad, lon: u.rad, t_idx=None):
        """
        Trace field lines.

        Parameters
        ----------
        mas_output : psipy.model.MASOutput
            MAS model output. Must have all three magnetic field components
            available.
        r : astropy.units.Quantity
            Radial seed coordinates.
        lat : astropy.units.Quantity
            Latitude seed points. Must be same shape as ``r``.
        lon : astropy.units.Quantity
            Longitude seed points. Must be same shape as ``r``.
        t_idx : int, optional
            Time slice of the ``mas_output`` to trace through. Doesn't need to
            be specified if only one time step is present.
        """
        r = r.to_value(mas_output.get_runit())
        lat = lat.to_value(u.rad)
        lon = lon.to_value(u.rad)
        seeds = np.stack([lon, lat, r], axis=-1)
        vector_grid = self._vector_grid(mas_output, t_idx)
        return self._trace_from_grid(vector_grid, seeds)

    def _trace_from_grid(self, grid, seeds):
        from streamtracer import StreamTracer

        seeds = np.atleast_2d(seeds)
        if self.max_steps == "auto":
            max_steps = int(4 * len(grid.zcoords) / self.step_size)
        else:
            max_steps = self.max_steps

        # Normalize step size to radial cell size
        rcoords = grid.zcoords
        step_size = self.step_size * np.min(np.diff(rcoords))
        self.tracer = StreamTracer(max_steps, step_size)
        self.tracer.trace(seeds, grid)
        return FieldLines(self.tracer.xs)
